@@ -1,10 +1,10 @@
 /* Conversion of files between different charsets and surfaces.
-   Copyright  1990, 93, 94, 96-99, 2000, 2001 Free Software Foundation, Inc.
-   Contributed by Franois Pinard <pinard@iro.umontreal.ca>, 1988.
+   Copyright © 1990-2023 Free Software Foundation, Inc.
+   Contributed by François Pinard <pinard@iro.umontreal.ca>, 1988.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public License
-   as published by the Free Software Foundation; either version 2 of the
+   as published by the Free Software Foundation; either version 3 of the
    License, or (at your option) any later version.
 
    This library is distributed in the hope that it will be
@@ -13,9 +13,9 @@
    Library General Public License for more details.
 
    You should have received a copy of the GNU Library General Public
-   License along with the `recode' Library; see the file `COPYING.LIB'.
-   If not, write to the Free Software Foundation, Inc., 59 Temple Place -
-   Suite 330, Boston, MA 02111-1307, USA.  */
+   License along with the Recode Library; see the file `COPYING.LIB'.
+   If not, see <https://www.gnu.org/licenses/>.
+*/
 
 /* Extended definitions, for externally written steps.  */
 
@@ -24,27 +24,20 @@
 /* Internal typedefs, to supplement those in "recode.h".  */
 
 typedef struct recode_symbol *			RECODE_SYMBOL;
-typedef struct recode_option_list * 		RECODE_OPTION_LIST;
+typedef struct recode_option_list *		RECODE_OPTION_LIST;
 typedef struct recode_single *			RECODE_SINGLE;
 typedef struct recode_step *			RECODE_STEP;
 typedef struct recode_alias *			RECODE_ALIAS;
 typedef struct recode_subtask *			RECODE_SUBTASK;
 
 typedef const struct recode_option_list *	RECODE_CONST_OPTION_LIST;
-typedef const struct recode_outer * 		RECODE_CONST_OUTER;
+typedef const struct recode_outer *		RECODE_CONST_OUTER;
 typedef const struct recode_step *		RECODE_CONST_STEP;
 typedef const struct recode_alias *		RECODE_CONST_ALIAS;
 typedef const struct recode_task *		RECODE_CONST_TASK;
 
-/*---------------------------------------------------------.
-| Maintain maximum of ERROR and current error in SUBTASK.  |
-`---------------------------------------------------------*/
-
-#define SET_SUBTASK_ERROR(Error, Subtask) \
-  recode_if_nogo (Error, Subtask)
-
 /*--------------------------------------------------------------------------.
-| Return from SUBTASK with `false', if the failure level has been reached.  |
+| Return from SUBTASK, with `false' if the failure level has been reached.  |
 `--------------------------------------------------------------------------*/
 
 #define SUBTASK_RETURN(Subtask) \
@@ -118,8 +111,15 @@ struct recode_outer
     /* This flag asks the library to diagnose and abort itself if errors.  */
     bool auto_abort;
 
-    /* If new steps should automatically have reversibility for fallback.  */
-    bool auto_reversibility;
+    /* If the external `iconv' library should be initialized and used.  */
+    bool use_iconv;
+
+    /* If we should discard untranslatable input and return an error,
+       unless 'force' is set (see below).  */
+    bool strict_mapping;
+
+    /* If we should ignore untranslatable input altogether.  */
+    bool force;
 
     /* charset.c */
     /* --------- */
@@ -137,8 +137,8 @@ struct recode_outer
     unsigned number_of_symbols;
 
     /* Arrays of strings ready for argmatch.  */
-    char **argmatch_charset_array;
-    char **argmatch_surface_array;
+    char const **argmatch_charset_array;
+    char const **argmatch_surface_array;
     const char **realname_charset_array;
     const char **realname_surface_array;
 
@@ -154,9 +154,8 @@ struct recode_outer
 
     /* Preset charsets and surfaces.  */
     RECODE_SYMBOL data_symbol;/* special charset defining surfaces */
-    RECODE_SYMBOL tree_symbol; /* special charset defining structures */
     RECODE_SYMBOL ucs2_charset; /* UCS-2 */
-    RECODE_SYMBOL libiconv_pivot; /* `libiconv' internal UCS */
+    RECODE_SYMBOL iconv_pivot; /* `iconv' internal UCS */
     RECODE_SYMBOL crlf_surface; /* for IBM PC machines */
     RECODE_SYMBOL cr_surface;	/* for Macintosh machines */
 
@@ -180,8 +179,7 @@ enum recode_symbol_type
   {
     RECODE_NO_SYMBOL_TYPE,	/* missing value */
     RECODE_CHARSET,		/* visible in the space of charsets */
-    RECODE_DATA_SURFACE,	/* this is a mere data surface */
-    RECODE_TREE_SURFACE		/* this is a structural surface */
+    RECODE_DATA_SURFACE		/* this is a mere data surface */
   };
 
 enum recode_data_type
@@ -202,6 +200,9 @@ struct recode_symbol
     /* Main name.  */
     const char *name;
 
+    /* Name to use for the iconv library, or NULL.  */
+    const char *iconv_name;
+
     /* Type of table.  */
     enum recode_data_type data_type;
 
@@ -231,12 +232,14 @@ struct recode_surface_list
 | Description of a single step of recoding.  |
 `-------------------------------------------*/
 
-typedef bool (*Recode_init) PARAMS ((RECODE_STEP, RECODE_CONST_REQUEST,
-				     RECODE_CONST_OPTION_LIST,
-				     RECODE_CONST_OPTION_LIST));
-typedef bool (*Recode_term) PARAMS ((RECODE_STEP, RECODE_CONST_REQUEST));
-typedef bool (*Recode_transform) PARAMS ((RECODE_SUBTASK));
-typedef bool (*Recode_fallback) PARAMS ((RECODE_SUBTASK, unsigned));
+extern bool recode_interrupted;
+
+typedef bool (*Recode_init) (RECODE_STEP, RECODE_CONST_REQUEST,
+                             RECODE_CONST_OPTION_LIST,
+                             RECODE_CONST_OPTION_LIST);
+typedef bool (*Recode_term) (RECODE_STEP);
+typedef bool (*Recode_transform) (RECODE_SUBTASK);
+typedef bool (*Recode_fallback) (RECODE_SUBTASK, unsigned);
 
 /* The `single' structure holds data needed to decide of sequences, and is
    invariant over actual requests.  The `step' structure holds data needed for
@@ -301,6 +304,7 @@ struct recode_step
 
     /* Recoding table.  */
     void *step_table;
+    void (*step_table_term_routine)(void *);
 
     /* Step specific variables.  */
     void *local;
@@ -348,7 +352,7 @@ struct recode_request
        to a single step, table production will fail.  */
     bool make_header_flag : 1;
 
-    /* For `latex' charset, it is often convenient to recode the diacritics
+    /* For `latex/bibtex' charsets, it is often convenient to recode the diacritics
        only, while letting other LaTeX code using backslashes unrecoded.  In
        the other charset, one can edit text as well as LaTeX directives.  */
     bool diacritics_only : 1;
@@ -379,7 +383,7 @@ struct recode_request
    BUFFER.  Moreover, if NAME is not NULL at start of recoding, this is
    interpreted as a request for the library to open the named file, either
    in read or write mode, and also to close it afterwards.  Standard input
-   or output is denoted by NAME being non-NULL, but otherwise empty.
+   or output is denoted by NAME being an empty string.
 
    If FILE is NULL in input mode, the in-memory read-only text extends from
    BUFFER to LIMIT.  There is no clue about if the buffer has been allocated
@@ -406,16 +410,6 @@ struct recode_read_write_text
     char *limit;
   };
 
-/* Tells how various passes are interconnected.  */
-
-enum recode_sequence_strategy
-  {
-    RECODE_STRATEGY_UNDECIDED,	/* sequencing strategy is undecided yet */
-    RECODE_SEQUENCE_IN_MEMORY,	/* keep intermediate recodings in memory */
-    RECODE_SEQUENCE_WITH_FILES,	/* do not fork, use intermediate files */
-    RECODE_SEQUENCE_WITH_PIPE	/* fork processes connected with `pipe(2)' */
-  };
-
 /* Tells how to swap the incoming pair of bytes, while reading UCS-2.  */
 
 enum recode_swap_input
@@ -468,9 +462,6 @@ struct recode_task
     struct recode_read_only_text input;
     struct recode_read_write_text output;
 
-    /* Tells how various recoding steps (passes) will be interconnected.  */
-    enum recode_sequence_strategy strategy : 3;
-
     /* Produce a byte order mark on UCS-2 output, insist for it on input.  */
     bool byte_order_mark : 1;
 
@@ -519,7 +510,7 @@ typedef unsigned short recode_ucs2;
    and 32 needs 25568, so we decided to stick to a strip size of 8.  Change
    $STRIP_SIZE in `doc/tables.pl' if you change the value here.  */
 
-/* "Are we speaking slips, strips or bars?" (of gold press'latinum :-) */
+/* "Are we speaking slips, strips or bars?" (of gold-pressed latinum :-) */
 #define STRIP_SIZE 8
 
 /* An struct strip_data is a pointer to a pool of strips, and an array
@@ -532,12 +523,6 @@ struct strip_data
     const short offset[256 / STRIP_SIZE];
   };
 
-struct ucs2_to_byte
-  {
-    recode_ucs2 code;		/* UCS-2 value */
-    unsigned char byte;		/* corresponding byte */
-  };
-
 struct ucs2_to_string
   {
     recode_ucs2 code;		/* UCS-2 value */
@@ -554,7 +539,7 @@ extern "C" {
 /* recode.c.  */
 
 #define ALLOC_SIZE(Variable, Size, Type) \
-  (Variable = (Type *) recode_malloc (outer, (Size)), Variable)
+  (Variable = (Type *) recode_malloc (outer, (Size)))
 
 #define ALLOC(Variable, Count, Type) \
   ALLOC_SIZE (Variable, (Count) * sizeof (Type), Type)
@@ -564,25 +549,25 @@ extern "C" {
 				       (Count) * sizeof(Type)), \
    Variable)
 
-void recode_error PARAMS ((RECODE_OUTER, const char *, ...));
-void recode_perror PARAMS ((RECODE_OUTER, const char *, ...));
-void *recode_malloc PARAMS ((RECODE_OUTER, size_t));
-void *recode_realloc PARAMS ((RECODE_OUTER, void *, size_t));
+void recode_error (RECODE_OUTER, const char *, ...);
+void recode_perror (RECODE_OUTER, const char *, ...);
+void *recode_malloc (RECODE_OUTER, size_t);
+void *recode_realloc (RECODE_OUTER, void *, size_t);
 
-unsigned char *invert_table PARAMS ((RECODE_OUTER, const unsigned char *));
-bool complete_pairs PARAMS ((RECODE_OUTER, RECODE_STEP,
-			     const struct recode_known_pair *, unsigned,
-			     bool, bool));
-bool transform_byte_to_ucs2 PARAMS ((RECODE_SUBTASK));
-bool init_ucs2_to_byte PARAMS ((RECODE_STEP, RECODE_CONST_REQUEST,
-				RECODE_CONST_OPTION_LIST,
-				RECODE_CONST_OPTION_LIST));
-bool transform_ucs2_to_byte PARAMS ((RECODE_SUBTASK));
+unsigned char *recode_invert_table (RECODE_OUTER, const unsigned char *);
+bool recode_complete_pairs (RECODE_OUTER, RECODE_STEP,
+                     const struct recode_known_pair *, unsigned,
+                     bool, bool);
+bool recode_transform_byte_to_ucs2 (RECODE_SUBTASK);
+bool recode_init_ucs2_to_byte (RECODE_STEP, RECODE_CONST_REQUEST,
+                               RECODE_CONST_OPTION_LIST,
+                               RECODE_CONST_OPTION_LIST);
+bool recode_transform_ucs2_to_byte (RECODE_SUBTASK);
 
 /* charname.c and fr-charname.c.  */
 
-const char *ucs2_to_charname PARAMS ((int));
-const char *ucs2_to_french_charname PARAMS ((int));
+const char *recode_ucs2_to_charname (int);
+const char *recode_ucs2_to_french_charname (int);
 
 /* charset.c.  */
 
@@ -590,72 +575,78 @@ enum alias_find_type
 {
   SYMBOL_CREATE_CHARSET,	/* charset as given, create as needed */
   SYMBOL_CREATE_DATA_SURFACE,	/* data surface as given, create as needed */
-  SYMBOL_CREATE_TREE_SURFACE,	/* tree surface as given, create as needed */
   ALIAS_FIND_AS_CHARSET,	/* disambiguate only as a charset */
   ALIAS_FIND_AS_SURFACE,	/* disambiguate only as a surface */
   ALIAS_FIND_AS_EITHER		/* disambiguate as a charset or a surface */
 };
 
-int code_to_ucs2 (RECODE_CONST_SYMBOL, unsigned);
-bool prepare_for_aliases PARAMS ((RECODE_OUTER));
-RECODE_ALIAS declare_alias PARAMS ((RECODE_OUTER,
-				     const char *, const char *));
-bool declare_implied_surface PARAMS ((RECODE_OUTER, RECODE_ALIAS,
-				    RECODE_CONST_SYMBOL));
-bool make_argmatch_arrays PARAMS ((RECODE_OUTER));
-RECODE_ALIAS find_alias PARAMS ((RECODE_OUTER, const char *,
-				   enum alias_find_type));
-bool find_and_report_subsets PARAMS ((RECODE_OUTER));
-bool decode_known_pairs PARAMS ((RECODE_OUTER, const char *));
+int recode_code_to_ucs2 (RECODE_CONST_SYMBOL, unsigned);
+bool recode_prepare_for_aliases (RECODE_OUTER);
+RECODE_ALIAS recode_declare_alias (RECODE_OUTER,
+                                   const char *, const char *);
+bool recode_declare_implied_surface (RECODE_OUTER, RECODE_ALIAS,
+                                     RECODE_CONST_SYMBOL);
+bool recode_make_argmatch_arrays (RECODE_OUTER);
+void recode_delete_alias (RECODE_ALIAS alias);
+RECODE_ALIAS recode_find_alias (RECODE_OUTER, const char *,
+                                enum alias_find_type);
+bool recode_find_and_report_subsets (RECODE_OUTER);
+bool recode_decode_known_pairs (RECODE_OUTER, const char *);
 
 /* combine.c.  */
 
 #define DONE NOT_A_CHARACTER
 #define ELSE BYTE_ORDER_MARK_SWAPPED
 
-bool init_explode PARAMS ((RECODE_STEP, RECODE_CONST_REQUEST,
-			   RECODE_CONST_OPTION_LIST,
-			   RECODE_CONST_OPTION_LIST));
-bool explode_byte_byte PARAMS ((RECODE_SUBTASK));
-bool explode_ucs2_byte PARAMS ((RECODE_SUBTASK));
-bool explode_byte_ucs2 PARAMS ((RECODE_SUBTASK));
-bool explode_ucs2_ucs2 PARAMS ((RECODE_SUBTASK));
+bool recode_init_explode (RECODE_STEP, RECODE_CONST_REQUEST,
+                          RECODE_CONST_OPTION_LIST,
+                          RECODE_CONST_OPTION_LIST);
+bool recode_explode_byte_byte (RECODE_SUBTASK);
+bool recode_explode_ucs2_byte (RECODE_SUBTASK);
+bool recode_explode_byte_ucs2 (RECODE_SUBTASK);
+bool recode_explode_ucs2_ucs2 (RECODE_SUBTASK);
 
-bool init_combine PARAMS ((RECODE_STEP, RECODE_CONST_REQUEST,
-			   RECODE_CONST_OPTION_LIST,
-			   RECODE_CONST_OPTION_LIST));
-bool combine_byte_byte PARAMS ((RECODE_SUBTASK));
-bool combine_ucs2_byte PARAMS ((RECODE_SUBTASK));
-bool combine_byte_ucs2 PARAMS ((RECODE_SUBTASK));
-bool combine_ucs2_ucs2 PARAMS ((RECODE_SUBTASK));
+bool recode_init_combine (RECODE_STEP, RECODE_CONST_REQUEST,
+                          RECODE_CONST_OPTION_LIST,
+                          RECODE_CONST_OPTION_LIST);
+bool recode_combine_byte_byte (RECODE_SUBTASK);
+bool recode_combine_ucs2_byte (RECODE_SUBTASK);
+bool recode_combine_byte_ucs2 (RECODE_SUBTASK);
+bool recode_combine_ucs2_ucs2 (RECODE_SUBTASK);
 
 /* freeze.c.  */
 
-void recode_freeze_tables PARAMS ((RECODE_OUTER));
+void recode_freeze_tables (RECODE_OUTER);
 
-/* libiconv.c.  */
+/* iconv.c.  */
 
-bool transform_with_libiconv PARAMS ((RECODE_SUBTASK));
+bool module_iconv (struct recode_outer *);
+void delmodule_iconv (struct recode_outer *);
+bool recode_transform_with_iconv (RECODE_SUBTASK);
+
+/* names.c.  */
+
+bool recode_should_prefer_french (void);
 
 /* mixed.c.  */
 
-bool transform_c_source PARAMS ((RECODE_TASK));
-bool transform_po_source PARAMS ((RECODE_TASK));
+bool recode_transform_c_source (RECODE_TASK);
+bool recode_transform_po_source (RECODE_TASK);
 
 /* outer.c.  */
 
-bool reversibility PARAMS ((RECODE_SUBTASK, unsigned));
-RECODE_SINGLE declare_single
-  PARAMS ((RECODE_OUTER, const char *, const char *,
-	   struct recode_quality,
-	   bool (*) (RECODE_STEP, RECODE_CONST_REQUEST,
-		     RECODE_CONST_OPTION_LIST, RECODE_CONST_OPTION_LIST),
-	   bool (*) (RECODE_SUBTASK)));
-bool declare_libiconv PARAMS ((RECODE_OUTER, const char *));
-bool declare_explode_data PARAMS ((RECODE_OUTER, const unsigned short *,
-				   const char *, const char *));
-bool declare_strip_data PARAMS ((RECODE_OUTER, struct strip_data *,
-				 const char *));
+bool recode_reversibility (RECODE_SUBTASK, unsigned);
+RECODE_SINGLE recode_declare_single
+  (RECODE_OUTER, const char *, const char *,
+   struct recode_quality,
+   bool (*) (RECODE_STEP, RECODE_CONST_REQUEST,
+             RECODE_CONST_OPTION_LIST, RECODE_CONST_OPTION_LIST),
+   bool (*) (RECODE_SUBTASK));
+bool recode_declare_iconv (RECODE_OUTER, const char *, const char *);
+bool recode_declare_explode_data (RECODE_OUTER, const unsigned short *,
+                                  const char *, const char *);
+bool recode_declare_strip_data (RECODE_OUTER, struct strip_data *,
+                                const char *);
 
 /* pool.c.  */
 
@@ -663,21 +654,21 @@ extern const recode_ucs2 ucs2_data_pool[];
 
 /* request.c.  */
 
-char *edit_sequence PARAMS ((RECODE_REQUEST, bool));
+char *recode_edit_sequence (RECODE_REQUEST, bool);
 
 /* rfc1345.c.  */
 
-const char *ucs2_to_rfc1345 PARAMS ((recode_ucs2));
+const char *recode_ucs2_to_rfc1345 (recode_ucs2);
 
 /* task.c.  */
 
-#if USE_HELPERS
-int get_byte_helper PARAMS ((RECODE_SUBTASK));
-#endif
-void put_byte_helper PARAMS ((int, RECODE_SUBTASK));
-bool recode_if_nogo PARAMS ((enum recode_error, RECODE_SUBTASK));
-bool transform_byte_to_byte PARAMS ((RECODE_SUBTASK));
-bool transform_byte_to_variable PARAMS ((RECODE_SUBTASK));
+int recode_get_byte (RECODE_SUBTASK);
+size_t recode_get_bytes (RECODE_SUBTASK, char *, size_t);
+void recode_put_byte (char, RECODE_SUBTASK);
+void recode_put_bytes (const char *, size_t, RECODE_SUBTASK);
+bool recode_if_nogo (enum recode_error, RECODE_SUBTASK);
+bool recode_transform_byte_to_byte (RECODE_SUBTASK);
+bool recode_transform_byte_to_variable (RECODE_SUBTASK);
 
 /* ucs.c.  */
 
@@ -685,7 +676,7 @@ bool transform_byte_to_variable PARAMS ((RECODE_SUBTASK));
    equivalent.  It is not used for ill-formed characters, however.  */
 #define REPLACEMENT_CHARACTER 0xFFFD
 
-/* Device for detetcing if bytes are swapped.  This value should appear first
+/* Device for detecting if bytes are swapped.  This value should appear first
    in UCS-2 files.  */
 #define BYTE_ORDER_MARK 0xFEFF
 #define BYTE_ORDER_MARK_SWAPPED 0xFFFE
@@ -693,61 +684,27 @@ bool transform_byte_to_variable PARAMS ((RECODE_SUBTASK));
 /* Never an UCS-2 character.  */
 #define NOT_A_CHARACTER 0xFFFF
 
-bool get_ucs2 PARAMS ((unsigned *, RECODE_SUBTASK));
-bool get_ucs4 PARAMS ((unsigned *, RECODE_SUBTASK));
-bool put_ucs2 PARAMS ((unsigned, RECODE_SUBTASK));
-bool put_ucs4 PARAMS ((unsigned, RECODE_SUBTASK));
+bool recode_get_ucs2 (unsigned *, RECODE_SUBTASK);
+bool recode_get_ucs4 (unsigned *, RECODE_SUBTASK);
+bool recode_put_ucs2 (unsigned, RECODE_SUBTASK);
+bool recode_put_ucs4 (unsigned, RECODE_SUBTASK);
 
 #ifdef __cplusplus
 }
 #endif
 
-/* Global macros specifically for `recode'.  */
+/* Global macros specifically for Recode.  */
 
 /* Giving a name to the ASCII character assigned to position 0.  */
 #define NUL '\0'
 
-#if USE_HELPERS
-
-# define get_byte(Subtask) \
-    get_byte_helper ((Subtask))
-
-# define put_byte(Byte, Subtask) \
-    put_byte_helper ((Byte), (Subtask))
-
-#else /* not USE_HELPERS */
-
-# define get_byte(Subtask) \
-    ((Subtask)->input.file					\
-     ? getc ((Subtask)->input.file)				\
-     : (Subtask)->input.cursor == (Subtask)->input.limit	\
-     ? EOF							\
-     : (unsigned char) *(Subtask)->input.cursor++)
-
-# define put_byte(Byte, Subtask) \
-    ((Subtask)->output.file					\
-     ? (putc ((char) (Byte), (Subtask)->output.file), 0)	\
-     : (Subtask)->output.cursor == (Subtask)->output.limit	\
-     ? (put_byte_helper ((int) (Byte), (Subtask)), 0)		\
-     : (*(Subtask)->output.cursor++ = (Byte), 0))
-
-#endif /* not USE_HELPERS */
-
 #ifdef FLEX_SCANNER
 
-# if !INLINE_HARDER
-
-#  undef put_byte
-#  define put_byte(Byte, Subtask) \
-     put_byte_helper ((Byte), (Subtask))
-
-# endif
-
 # define PUT_NON_DIACRITIC_BYTE(Byte, Subtask) \
     if (request->diacritics_only)				\
       ECHO;							\
     else							\
-      put_byte ((Byte), (Subtask))
+      recode_put_byte ((Byte), (Subtask))
 
 /* ECHO may not have a (Subtask) argument, because some ECHO without argument
    is generated by Flex -- yet Vern tells me it won't happen if I inhibit
@@ -755,9 +712,10 @@ bool put_ucs4 PARAMS ((unsigned, RECODE_SUBTASK));
    `subtask' quite systematically, so it may be used as a constant, here.  */
 # define ECHO \
     do {							\
-      const char *cursor = yytext; int counter = yyleng;	\
+      const char *cursor = librecode_yytext;			\
+      int counter = librecode_yyleng;				\
       for (; counter > 0; cursor++, counter--)			\
-	put_byte (*cursor, subtask);				\
+	recode_put_byte (*cursor, subtask);				\
     } while (false)
 
 #endif /* FLEX_SCANNER */
